# -*- coding: utf-8 -*-

import logging

import numpy as np
import torch as th
import torch.nn as nn

from leibniz.nn.conv import DepthwiseSeparableConv1d, DepthwiseSeparableConv2d, DepthwiseSeparableConv3d
from leibniz.nn.layer.cbam import CBAM
from leibniz.nn.net.hyptube import HypTube
from leibniz.nn.net.unet import UNet, Enconv, Deconv, Block, Transform

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class UNet2(nn.Module):
    def __init__(self, in_channels, out_channels, block=None, attn=CBAM, relu=None, layers=4, ratio=2, enhencer=None, ksize_in=7, dropout_prob=0.1,
                 vblks=None, hblks=None, scales=None, factors=None, spatial=(256, 256), normalizor='batch', padding=None, final_normalized=True):
        super().__init__()

        extension = block.extension
        lrd = block.least_required_dim

        spatial = np.array(spatial, dtype=np.int)
        dim = len(spatial)
        self.dim = dim
        Conv = self.get_conv_for_prepare()
        TConv = self.get_conv_for_transform()

        scales = np.array(scales)
        if scales.shape[0] != layers:
            raise ValueError('scales should have %d layers at dim 0!' % layers)
        if len(scales.shape) == 1:
            scales = scales.reshape([layers, 1])
        if len(scales.shape) != 2:
            raise ValueError('scales should have length 2 to be compatible with spatial dimensions!')

        ratio = np.exp2(ratio)
        factors = np.array(factors + [0.0])
        scales = np.exp2(scales)
        factors = np.exp2(factors)
        num_filters = int(in_channels * ratio)

        self.final_normalized = final_normalized
        self.ratio = ratio
        self.hblks = hblks
        self.vblks = vblks
        self.scales = scales
        self.factors = factors
        logger.info('---------------------------------------')
        logger.info('ratio: %f', ratio)
        logger.info('vblks: [%s]', ', '.join(map(str, vblks)))
        logger.info('hblks: [%s]', ', '.join(map(str, hblks)))
        logger.info('scales: [%s]', ', '.join(map(str, scales)))
        logger.info('factors: [%s]', ', '.join(map(str, factors[0:4])))
        logger.info('---------------------------------------')

        self.exceeded = np.any(np.cumprod(scales, axis=0) * spatial < 1) or np.any((in_channels * ratio * np.cumprod(factors)) < lrd)
        if not self.exceeded:
            self.layers = layers
            self.in_channels = in_channels
            self.num_filters = num_filters
            self.out_channels = out_channels

            if enhencer is None:
                if self.dim == 2:
                    enhencer = HypTube
            self.enhencer_in = None
            self.enhencer_out = None
            self.enhencer_mid = None

            if relu is None:
                relu = nn.ReLU(inplace=True)

            if attn is None:
                attn = CBAM

            ex = extension
            c0 = int(ex * num_filters)
            if padding:
                self.conv_padding = 0
                self.iconv = nn.Sequential(
                    padding,
                    Conv(in_channels, c0, kernel_size=ksize_in, padding=(ksize_in - 1) // 2, groups=1),
                )
                self.oconv = nn.Sequential(
                    padding,
                    Conv(c0, out_channels, kernel_size=3, padding=self.conv_padding, bias=False, groups=1),
                )
            else:
                self.conv_padding = 1
                self.iconv = Conv(in_channels, c0, kernel_size=ksize_in, padding=(ksize_in - 1) // 2, groups=1)
                self.oconv = Conv(c0 * 2, out_channels, kernel_size=3, padding=self.conv_padding, bias=False, groups=1)

            if final_normalized:
                self.relu6 = nn.ReLU6()

            self.enconvs = nn.ModuleList()
            self.dnforms = nn.ModuleList()
            self.hzforms = nn.ModuleList()
            self.upforms = nn.ModuleList()
            self.deconvs = nn.ModuleList()
            self.mdconvs = nn.ModuleList()

            self.spatial = [np.array(spatial, dtype=np.int)]
            self.channel_sizes = [np.array(c0, dtype=np.int)]
            for ix in range(layers):
                least_factor = ex
                scale, factor = scales[ix], factors[ix]
                self.spatial.append(np.array(self.spatial[ix] * scale, dtype=np.int))
                self.channel_sizes.append(np.array(self.channel_sizes[ix] * factor // least_factor * least_factor, dtype=np.int))

                ci, co = self.channel_sizes[ix].item(), self.channel_sizes[ix + 1].item()
                szi, szo = self.spatial[ix + 1], self.spatial[ix]
                logger.info('%d - ci: %d, co: %d', ix, ci, co)
                logger.info('%d - szi: [%s], szo: [%s]', ix, ', '.join(map(str, szi)), ', '.join(map(str, szo)))

                self.exceeded = self.exceeded or ci < lrd or co < lrd or szi.min() < 1 or szo.min() < 1
                if not self.exceeded:
                    try:
                        dropout_flag = (layers - ix) * 3 < layers
                        dropout = dropout_prob if dropout_flag else -1
                        self.mdconvs.append(Block(Deconv(co, ci, size=szo, conv=TConv, padding=padding), activation=True, dropout=False,
                                                  relu=relu, attn=attn, dim=self.dim, normalizor=normalizor, conv=TConv))
                        self.enconvs.append(Block(Enconv(ci, co, size=szi, conv=TConv, padding=padding), activation=True, dropout=dropout,
                                                  relu=relu, attn=attn, dim=self.dim, normalizor=normalizor, conv=TConv))
                        self.dnforms.append(Transform(co, co, nblks=vblks[ix], block=block, relu=relu, conv=TConv))
                        self.hzforms.append(Transform(co * 2, co * 2, nblks=hblks[ix], block=block, relu=relu, conv=TConv))
                        self.deconvs.append(Block(Deconv(co * 3, ci, size=szo, conv=TConv, padding=padding), activation=True, dropout=False,
                                                  relu=relu, attn=attn, dim=self.dim, normalizor=normalizor, conv=TConv))
                        self.upforms.append(Transform(ci, ci, nblks=vblks[ix], block=block, relu=relu, conv=TConv))
                    except Exception as e:
                        logger.exception(e)
                        self.exceeded = True
                else:
                    logger.error('scales are exceeded!')
                    raise ValueError('scales exceeded!')

            if self.dim == 2 and enhencer is not None:
                self.enhencer_mid = enhencer(co, (c0 + 1) // 2, co)

    def get_conv_for_prepare(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError('dim %d is not supported!' % self.dim)
        return conv

    def get_conv_for_transform(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError('dim %d is not supported!' % self.dim)
        return conv

    def forward(self, x):
        if self.exceeded:
            raise ValueError('scales exceeded!')

        dnt = self.iconv(x)

        dnts, encs, mids = [], [], []
        for ix in range(self.layers):
            dnt, enc = self.dnforms[ix](self.enconvs[ix](dnt))
            mid = self.mdconvs[ix](enc)
            dnts.append(dnt)
            encs.append(enc)
            mids.append(mid)

        if self.enhencer_mid is not None:
            upt = self.enhencer_mid(dnt)
        else:
            upt = dnt

        for ix in range(self.layers - 1, -1, -1):
            enc = encs[ix]
            if ix == self.layers - 1:
                mid = enc

            inp = th.cat([enc, mid], dim=1)
            trans = self.hzforms[ix]
            hzt, _ = trans(inp)
            upt, dec = self.upforms[ix](self.deconvs[ix](upt, hzt))
            mid = mids[ix]

        return self.oconv(th.cat([upt, mid], dim=1))



import torch.nn as nn


class MLP1d(nn.Module):
    def __init__(self, channels_in, channels_out):
        super(MLP1d, self).__init__()
        channels_hidden = channels_in + channels_out
        self.layers = nn.Sequential(
            nn.Linear(channels_in, channels_hidden),
            nn.ReLU(),
            nn.Linear(channels_hidden, channels_out)
        )

    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = self.layers(x)
        return x




#
#  Copy from https://github.com/Jimexist/conv_lstm_pytorch/blob/master/conv_lstm/conv_lstm.py
#  Originally by Jimexist under MIT License
#

import torch
import torch.nn as nn

class ConvLSTMCell(nn.Module):
    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        kernel_size: int,
        bias: bool,
    ):
        """
        Initialize ConvLSTM cell.
        Parameters
        ----------
        input_dim: int
            Number of channels of input tensor.
        hidden_dim: int
            Number of channels of hidden state.
        kernel_size: (int, int)
            Size of the convolutional kernel.
        bias: bool
            Whether or not to add the bias.
        """

        super().__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.kernel_size = (kernel_size, kernel_size)
        self.padding = self.kernel_size[0] // 2, self.kernel_size[1] // 2
        self.bias = bias

        self.conv = nn.Conv2d(
            in_channels=self.input_dim + self.hidden_dim,
            out_channels=4 * self.hidden_dim,
            kernel_size=self.kernel_size,
            stride=1,
            padding=self.padding,
            bias=self.bias,
        )

    def forward(self, input_tensor, cur_state):
        h_cur, c_cur = cur_state

        # concatenate along channel axis
        combined = torch.cat([input_tensor, h_cur], dim=1)

        combined_conv = self.conv(combined)
        cc_i, cc_f, cc_o, cc_g = torch.split(combined_conv, self.hidden_dim, dim=1)
        i = torch.sigmoid(cc_i)
        f = torch.sigmoid(cc_f)
        o = torch.sigmoid(cc_o)
        g = torch.tanh(cc_g)

        c_next = f * c_cur + i * g
        h_next = o * torch.tanh(c_next)

        return h_next, c_next

    def init_hidden(self, batch_size, image_size):
        height, width = image_size
        return (
            torch.zeros(
                batch_size,
                self.hidden_dim,
                height,
                width,
                device=self.conv.weight.device,
            ),
        ) * 2


class ConvLSTM(nn.Module):
    """
    Parameters:
        input_dim: Number of channels in input
        hidden_dim: Number of hidden channels
        kernel_size: Size of kernel in convolutions
        num_layers: Number of LSTM layers stacked on each other
        batch_first: Whether or not dimension 0 is the batch or not
        bias: Bias or no bias in Convolution
        return_all_layers: Return the list of computations for all layers
        Note: Will do same padding.
    Input:
        A tensor of size B, T, C, H, W or T, B, C, H, W
    Output:
        A tuple of two lists of length num_layers (or length 1 if return_all_layers is False).
            0 - layer_output_list is the list of lists of length T of each output
            1 - last_state_list is the list of last states
                    each element of the list is a tuple (h, c) for hidden state and memory
    Example:
        >> x = torch.rand((32, 10, 64, 128, 128))
        >> convlstm = ConvLSTM(64, 16, 3, 1, True, True, False)
        >> _, last_states = convlstm(x)
        >> h = last_states[0][0]  # 0 for layer index, 0 for h index
    """

    def __init__(
        self,
        input_dim,
        hidden_dim,
        kernel_size,
        num_layers=1,
        batch_first=True,
        bias=True,
        return_all_layers=False,
    ):
        super().__init__()
        # Make sure that both `kernel_size` and `hidden_dim` are lists having len == num_layers
        kernel_size = self._extend_for_multilayer(kernel_size, num_layers)
        hidden_dim = self._extend_for_multilayer(hidden_dim, num_layers)
        assert (
            len(kernel_size) == len(hidden_dim) == num_layers
        ), "Inconsistent list length."

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.kernel_size = kernel_size
        self.num_layers = num_layers
        self.batch_first = batch_first
        self.bias = bias
        self.return_all_layers = return_all_layers

        cell_list = []
        for i in range(self.num_layers):
            cur_input_dim = self.input_dim if i == 0 else self.hidden_dim[i - 1]

            cell_list.append(
                ConvLSTMCell(
                    input_dim=cur_input_dim,
                    hidden_dim=self.hidden_dim[i],
                    kernel_size=self.kernel_size[i],
                    bias=self.bias,
                )
            )

        self.cell_list = nn.ModuleList(cell_list)

    def forward(self, input_tensor):
        """
        Parameters
        ----------
        input_tensor: todo
            5-D Tensor either of shape (t, b, c, h, w) or (b, t, c, h, w)
        Returns
        -------
        last_state_list, layer_output
        """
        if not self.batch_first:
            # (t, b, c, h, w) -> (b, t, c, h, w)
            input_tensor = input_tensor.permute(1, 0, 2, 3, 4)

        b, _, _, h, w = input_tensor.size()

        # Since the init is done in forward. Can send image size here
        hidden_state = self._init_hidden(batch_size=b, image_size=(h, w))

        layer_output_list = []
        last_state_list = []

        seq_len = input_tensor.size(1)
        cur_layer_input = input_tensor

        for layer_idx in range(self.num_layers):
            h, c = hidden_state[layer_idx]
            output_inner = []
            for t in range(seq_len):
                h, c = self.cell_list[layer_idx](
                    input_tensor=cur_layer_input[:, t, :, :, :], cur_state=[h, c]
                )
                output_inner.append(h)

            layer_output = torch.stack(output_inner, dim=1)
            cur_layer_input = layer_output

            layer_output_list.append(layer_output)
            last_state_list.append([h, c])

        if not self.return_all_layers:
            layer_output_list = layer_output_list[-1:]
            last_state_list = last_state_list[-1:]

        return layer_output_list, last_state_list

    def _init_hidden(self, batch_size, image_size):
        init_states = []
        for i in range(self.num_layers):
            init_states.append(self.cell_list[i].init_hidden(batch_size, image_size))
        return init_states

    @staticmethod
    def _extend_for_multilayer(param, num_layers: int):
        if isinstance(param, int):
            return (param,) * num_layers
        elif isinstance(param, list):
            return tuple(param)
        else:
            return param


from torch import nn as nn


def conv7x7(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """7x7 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=7, stride=stride,
                     padding=3, groups=groups, bias=False, dilation=dilation)


def conv5x5(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """5x5 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=5, stride=stride,
                     padding=2, groups=groups, bias=False, dilation=dilation)


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class DepthwiseSeparableConv1d(nn.Module):
    def __init__(self, in_channels, output_channels, kernel_size=1, padding=0, stride=1, dilation=1, groups=1, bias=True, kernels_per_layer=1):
        super(DepthwiseSeparableConv1d, self).__init__()
        # In Tensorflow DepthwiseConv2D has depth_multiplier instead of kernels_per_layer
        self.depthwise = nn.Conv1d(in_channels, in_channels * kernels_per_layer, kernel_size, padding=padding,
                                   stride=stride, dilation=dilation, groups=in_channels, bias=bias)
        self.pointwise = nn.Conv1d(in_channels * kernels_per_layer, output_channels, 1, padding=0,
                                   stride=stride, dilation=dilation, groups=groups, bias=bias)

    def forward(self, x):
        x = self.depthwise(x)
        x = self.pointwise(x)
        return x


class DepthwiseSeparableConv2d(nn.Module):
    def __init__(self, in_channels, output_channels, kernel_size=1, padding=0, stride=1, dilation=1, groups=1, bias=True, kernels_per_layer=1):
        super(DepthwiseSeparableConv2d, self).__init__()
        # In Tensorflow DepthwiseConv2D has depth_multiplier instead of kernels_per_layer
        self.depthwise = nn.Conv2d(in_channels, in_channels * kernels_per_layer, kernel_size, padding=padding,
                                   stride=stride, dilation=dilation, groups=in_channels, bias=bias)
        self.pointwise = nn.Conv2d(in_channels * kernels_per_layer, output_channels, 1, padding=0,
                                   stride=stride, dilation=dilation, groups=groups, bias=bias)

    def forward(self, x):
        x = self.depthwise(x)
        x = self.pointwise(x)
        return x


class DepthwiseSeparableConv3d(nn.Module):
    def __init__(self, in_channels, output_channels, kernel_size=1, padding=0, stride=1, dilation=1, groups=1, bias=True, kernels_per_layer=1):
        super(DepthwiseSeparableConv3d, self).__init__()
        self.depthwise = nn.Conv3d(in_channels, in_channels * kernels_per_layer, kernel_size, padding=padding,
                                   stride=stride, dilation=dilation, groups=in_channels, bias=bias)
        self.pointwise = nn.Conv3d(in_channels * kernels_per_layer, output_channels, 1, padding=0,
                                   stride=stride, dilation=dilation, groups=groups, bias=bias)

    def forward(self, x):
        x = self.depthwise(x)
        x = self.pointwise(x)
        return x


class ComplexConv1d(nn.Module):
    def __init__(self,in_channels, out_channels, kernel_size=3, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(ComplexConv1d, self).__init__()
        self.rconv = nn.Conv1d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)
        self.iconv = nn.Conv1d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        return self.rconv(input.real) - self.iconv(input.imag) + 1j * (self.rconv(input.imag) + self.iconv(input.real))


class ComplexConv2d(nn.Module):
    def __init__(self,in_channels, out_channels, kernel_size=3, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(ComplexConv2d, self).__init__()
        self.rconv = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)
        self.iconv = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        return self.rconv(input.real) - self.iconv(input.imag) + 1j * (self.rconv(input.imag) + self.iconv(input.real))


class ComplexConv3d(nn.Module):
    def __init__(self,in_channels, out_channels, kernel_size=3, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(ComplexConv3d, self).__init__()
        self.rconv = nn.Conv3d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)
        self.iconv = nn.Conv3d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        return self.rconv(input.real) - self.iconv(input.imag) + 1j * (self.rconv(input.imag) + self.iconv(input.real))



# -*- coding: utf-8 -*-

import torch as th
import leibniz as lbnz

from cached_property import cached_property
from torch import Tensor
from typing import Tuple


class ThetaPhiRFrame:
    def __init__(self, grid):
        self.default_device = -1
        self.grid = grid
        lbnz.use('theta,phi,r')

        self.thx = - th.sin(lbnz.theta)
        self.thy = th.cos(lbnz.theta)
        self.thz = - lbnz.zero

        self.phx = - th.sin(lbnz.phi) * th.cos(lbnz.theta)
        self.phy = - th.sin(lbnz.phi) * th.sin(lbnz.theta)
        self.phz = th.cos(lbnz.phi)

        self.rx = th.cos(lbnz.phi) * th.cos(lbnz.theta)
        self.ry = th.cos(lbnz.phi) * th.sin(lbnz.theta)
        self.rz = th.sin(lbnz.phi)

    def get_device(self):
        return self.default_device

    def set_device(self, ix):
        self.thx = self.thx.cuda(device=ix)
        self.thy = self.thy.cuda(device=ix)
        self.thz = self.thz.cuda(device=ix)
        self.phx = self.phx.cuda(device=ix)
        self.phy = self.phy.cuda(device=ix)
        self.phz = self.phz.cuda(device=ix)
        self.rx = self.rx.cuda(device=ix)
        self.ry = self.ry.cuda(device=ix)
        self.rz = self.rz.cuda(device=ix)
        self.default_device = ix

    @cached_property
    def phi(self) -> Tuple[Tensor, Tensor, Tensor]:
        return self.phx, self.phy, self.phz

    @cached_property
    def theta(self) -> Tuple[Tensor, Tensor, Tensor]:
        return self.thx, self.thy, self.thz

    @cached_property
    def r(self) -> Tuple[Tensor, Tensor, Tensor]:
        return self.rx, self.ry, self.rz


_name_ = 'thetaphir'
_clazz_ = ThetaPhiRFrame



# -*- coding: utf-8 -*-

import torch as th


def add(x, y):
    x1, x2 = x
    y1, y2 = y

    return x1 + y1, x2 + y2


def sub(x, y):
    x1, x2 = x
    y1, y2 = y

    return x1 - y1, x2 - y2


def mult(x, y):
    x1, x2 = x
    y1, y2 = y

    return x1 * y1, x2 * y2


def div(x, y):
    x1, x2 = x
    y1, y2 = y

    return x1 / y1, x2 / y2


def dot(x, y):
    x1, x2 = x
    y1, y2 = y

    return x1 * y1 + x2 * y2


def cross(x, y):
    x1, x2 = x
    y1, y2 = y

    return x1 * y2 - x2 * y1


def norm(v):
    return th.sqrt(th.sum(dot(v, v), dim=1, keepdim=True))


def normsq(v):
    return th.sum(dot(v, v), dim=1, keepdim=True)


def normalize(v):
    x, y = v
    r = th.sqrt(x * x + y * y)

    return x / r, y / r


def det(transform):
    a, b, c, d = transform[0, 0], transform[0, 1], transform[1, 0], transform[1, 1]
    return a * d - b * c



# -*- coding: utf-8 -*-

import torch as th
import torch.nn as nn

from leibniz.nn.layer.simam import SimAM


class BasicBlock(nn.Module):
    def __init__(self, in_channel, out_channel, step, relu, conv, reduction=16):
        super(BasicBlock, self).__init__()
        self.step = step
        self.relu = relu

        self.conv1 = conv(in_channel, in_channel, kernel_size=3, stride=1, padding=1)
        self.conv2 = conv(in_channel, out_channel, kernel_size=3, stride=1, padding=1)
        self.simam = SimAM(out_channel, reduction)

    def forward(self, x):
        y = self.conv1(x)
        y = self.relu(y)
        y = self.conv2(y)
        y = self.simam(y)
        return y


class Bottleneck(nn.Module):
    def __init__(self, in_channel, out_channel, step, relu, conv, reduction=16):
        super(Bottleneck, self).__init__()
        self.step = step
        self.relu = relu
        hidden = min(in_channel, out_channel) // 4 + 1

        self.conv1 = conv(in_channel, hidden, kernel_size=1, bias=False)
        self.conv2 = conv(hidden, hidden, kernel_size=3, bias=False, padding=1)
        self.conv3 = conv(hidden, out_channel, kernel_size=1, bias=False)
        self.simam = SimAM(out_channel, reduction)

    def forward(self, x):
        y = self.conv1(x)
        y = self.relu(y)
        y = self.conv2(y)
        y = self.relu(y)
        y = self.conv3(y)
        y = self.simam(y)
        return y


class HyperBasic(nn.Module):
    extension = 1
    least_required_dim = 1

    def __init__(self, dim, step, ix, tx, relu, conv, reduction=16):
        super(HyperBasic, self).__init__()
        self.dim = dim
        self.step = step
        self.ix = ix
        self.tx = tx

        self.input = BasicBlock(dim, 2 * dim, step, relu, conv, reduction=reduction)
        self.output = BasicBlock(7 * dim, dim, step, relu, conv, reduction=reduction)

    def forward(self, x):
        r = self.input(x)
        velo = r[:, :self.dim]
        theta = r[:, self.dim:]
        u = velo * th.cos(theta)
        v = velo * th.sin(theta)

        y1 = x * (1 + v * self.step) + u * self.step
        y2 = x * (1 + u * self.step) - v * self.step
        y3 = x * (1 - v * self.step) - u * self.step
        y4 = x * (1 - u * self.step) + v * self.step
        ys = th.cat([y1, y2, y3, y4, x, velo, theta], dim=1)
        r = self.output(ys) * self.step

        return y1 * (1 + r)


class HyperBottleneck(nn.Module):
    extension = 2
    least_required_dim = 1

    def __init__(self, dim, step, ix, tx, relu, conv, reduction=16):
        super(HyperBottleneck, self).__init__()
        self.dim = dim
        self.step = step
        self.ix = ix
        self.tx = tx

        self.input = Bottleneck(dim, 2 * dim, step, relu, conv, reduction=reduction)
        self.output = Bottleneck(7 * dim, dim, step, relu, conv, reduction=reduction)

    def forward(self, x):
        r = self.input(x)
        velo = r[:, :self.dim]
        theta = r[:, self.dim:]
        u = velo * th.cos(theta)
        v = velo * th.sin(theta)

        y1 = x * (1 + v * self.step) + u * self.step
        y2 = x * (1 + u * self.step) - v * self.step
        y3 = x * (1 - v * self.step) - u * self.step
        y4 = x * (1 - u * self.step) + v * self.step
        ys = th.cat([y1, y2, y3, y4, x, velo, theta], dim=1)
        r = self.output(ys) * self.step

        return y1 * (1 + r)



import torch as th
import torch.nn as nn
import torch.nn.functional as F


class BilinearWarpingScheme(nn.Module):
    def __init__(self, padding_mode='zeros'):
        super(BilinearWarpingScheme, self).__init__()
        self.padding_mode = padding_mode
        self.grids = {}

    def forward(self, im, ws):
        b = ws.size()[0]
        c = ws.size()[1]
        h = ws.size()[2]
        w = ws.size()[3]

        key = '%d-%s' % (b, im.get_device())
        if key not in self.grids:
            if im.get_device() < 0:
                g0 = th.linspace(-1, 1, h, requires_grad=False)
                g1 = th.linspace(-1, 1, w, requires_grad=False)
            else:
                g0 = th.linspace(-1, 1, h, device=im.get_device(), requires_grad=False)
                g1 = th.linspace(-1, 1, w, device=im.get_device(), requires_grad=False)
            grid = th.cat(th.meshgrid([g0, g1]), dim=1).reshape(1, 2, h, w)
            self.grids[key] = grid.repeat(b * c // 2, 1, 1, 1)

        grid = self.grids[key]
        shift = grid.reshape(-1, 2, h, w) - ws.reshape(-1, 2, h, w)
        shift = shift.permute(0, 2, 3, 1)
        return F.grid_sample(im.reshape(-1, 2, h, w), shift, padding_mode=self.padding_mode, mode='bilinear').reshape(-1, c, h, w)



# -*- coding: utf-8 -*-

import logging

import numpy as np
import torch.nn as nn

from leibniz.nn.net.unet import Enconv, Transform, Block
from leibniz.nn.conv import DepthwiseSeparableConv1d, DepthwiseSeparableConv2d, DepthwiseSeparableConv3d
from leibniz.nn.layer.cbam import CBAM

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class ResNet(nn.Module):
    def __init__(self, in_channels, out_channels, block=None, attn=None, relu=None, layers=4, ratio=2, dropout_prob=0.5,
                 vblks=None, factors=None, spatial=(256, 256), normalizor='batch', padding=None, final_normalized=True):
        super().__init__()

        extension = block.extension
        lrd = block.least_required_dim

        spatial = np.array(spatial, dtype=np.int)
        dim = len(spatial)
        self.dim = dim
        Conv = self.get_conv_for_prepare()
        TConv = self.get_conv_for_transform()

        ratio = np.exp2(ratio)
        factors = np.array(factors + [0.0])
        factors = np.exp2(factors)
        num_filters = int(in_channels * ratio)

        self.final_normalized = final_normalized
        self.ratio = ratio
        self.vblks = vblks
        self.factors = factors
        logger.info('---------------------------------------')
        logger.info('ratio: %f', ratio)
        logger.info('vblks: [%s]', ', '.join(map(str, vblks)))
        logger.info('factors: [%s]', ', '.join(map(str, factors[0:4])))
        logger.info('---------------------------------------')

        self.layers = layers
        self.in_channels = in_channels
        self.num_filters = num_filters
        self.out_channels = out_channels

        if relu is None:
            relu = nn.ReLU(inplace=True)

        if attn is None:
            attn = CBAM

        ex = extension
        c0 = int(ex * num_filters)
        accumulated_factors = np.cumprod(factors, axis=0)[-1]
        accumulated_spatial = np.cumprod(spatial, axis=0)[-1]
        pn = int(c0 * accumulated_factors * accumulated_spatial)

        if padding:
            self.conv_padding = 0
            self.iconv = nn.Sequential(
                padding,
                Conv(in_channels, c0, kernel_size=3, padding=self.conv_padding, groups=1),
            )
            self.fc = nn.Sequential(
                padding,
                nn.Linear(pn, out_channels),
            )
        else:
            self.conv_padding = 1
            self.iconv = Conv(in_channels, c0, kernel_size=7, padding=3, groups=1)
            self.fc = nn.Linear(pn, out_channels, bias=False)

        if final_normalized:
            self.relu6 = nn.ReLU6()

        self.enconvs = nn.ModuleList()
        self.transforms = nn.ModuleList()

        self.spatial = [np.array(spatial, dtype=np.int)]
        self.channel_sizes = [np.array(c0, dtype=np.int)]
        for ix in range(layers):
            least_factor = ex
            factor = factors[ix]
            self.channel_sizes.append(np.array(self.channel_sizes[ix] * factor // least_factor * least_factor, dtype=np.int))

            ci, co = self.channel_sizes[ix].item(), self.channel_sizes[ix + 1].item()
            logger.info('%d - ci: %d, co: %d', ix, ci, co)

            self.enconvs.append(Block(Enconv(ci, co, size=spatial, conv=TConv, padding=padding), activation=True, dropout=dropout_prob, relu=relu, attn=attn, dim=self.dim, normalizor=normalizor, conv=TConv))
            self.transforms.append(Transform(co, co, nblks=vblks[ix], block=block, relu=relu, conv=TConv))

    def get_conv_for_prepare(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError('dim %d is not supported!' % self.dim)
        return conv

    def get_conv_for_transform(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError('dim %d is not supported!' % self.dim)
        return conv

    def forward(self, x):
        sz = x.size()

        dnt = self.iconv(x)
        for ix in range(self.layers):
            dnt, enc = self.transforms[ix](self.enconvs[ix](dnt))

        return self.fc(dnt.view(sz[0], -1))



import torch.nn as nn
import torch.nn.functional as F


class ComplexDropout1d(nn.Module):
    def __init__(self, p=0.5, inplace=True):
        super(ComplexDropout1d, self).__init__()
        self.p = p
        self.inplace = inplace

    def forward(self, input):
        return F.dropout(input.real, p=self.p, training=self.training, inplace=self.inplace) +\
               1j * F.dropout(input.imag, p=self.p, training=self.training, inplace=self.inplace)


class ComplexDropout2d(nn.Module):
    def __init__(self, p=0.5, inplace=True):
        super(ComplexDropout2d, self).__init__()
        self.p = p
        self.inplace = inplace

    def forward(self, input):
        return F.dropout2d(input.real, p=self.p, training=self.training, inplace=self.inplace) +\
               1j * F.dropout2d(input.imag, p=self.p, training=self.training, inplace=self.inplace)


class ComplexDropout3d(nn.Module):
    def __init__(self, p=0.5, inplace=True):
        super(ComplexDropout3d, self).__init__()
        self.p = p
        self.inplace = inplace

    def forward(self, input):
        return F.dropout3d(input.real, p=self.p, training=self.training, inplace=self.inplace) +\
               1j * F.dropout3d(input.imag, p=self.p, training=self.training, inplace=self.inplace)



import torch as th
import torch.nn as nn


# Originally from Kevin Trebing (HansBambel) at https://github.com/HansBambel/SmaAt-UNet/blob/master/models/layers.py
# SmaAt-UNet: Precipitation Nowcasting using a Small Attention-UNet Architecture
# https://arxiv.org/abs/2007.04417v1
# Modified by Mingli Yuan (Mountain)

class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size(0), -1)


class ChannelLayer(nn.Module):
    def __init__(self, input_channels, reduction_ratio=16):
        super(ChannelLayer, self).__init__()
        self.input_channels = input_channels
        self.avg_pool = None
        self.max_pool = None
        #  https://github.com/luuuyi/CBAM.PyTorch/blob/master/model/resnet_cbam.py
        #  uses Convolutions instead of Linear
        self.mlp = nn.Sequential(
            Flatten(),
            nn.Linear(input_channels, input_channels // reduction_ratio + 1),
            nn.ReLU(),
            nn.Linear(input_channels // reduction_ratio + 1, input_channels)
        )

    def forward(self, x):
        sz = x.size()
        if len(sz) == 3:
            if self.avg_pool is None:
                self.avg_pool = nn.AdaptiveAvgPool1d(1)
                self.max_pool = nn.AdaptiveMaxPool1d(1)
            # Take the input and apply average and max pooling
            avg_values = self.avg_pool(x)
            max_values = self.max_pool(x)
            out = (self.mlp(avg_values) + self.mlp(max_values)).view(sz[0], sz[1], 1)
        if len(sz) == 4:
            if self.avg_pool is None:
                self.avg_pool = nn.AdaptiveAvgPool2d(1)
                self.max_pool = nn.AdaptiveMaxPool2d(1)
            # Take the input and apply average and max pooling
            avg_values = self.avg_pool(x)
            max_values = self.max_pool(x)
            out = (self.mlp(avg_values) + self.mlp(max_values)).view(sz[0], sz[1], 1, 1)
        if len(sz) == 5:
            if self.avg_pool is None:
                self.avg_pool = nn.AdaptiveAvgPool3d(1)
                self.max_pool = nn.AdaptiveMaxPool3d(1)
            # Take the input and apply average and max pooling
            avg_values = self.avg_pool(x)
            max_values = self.max_pool(x)
            out = (self.mlp(avg_values) + self.mlp(max_values)).view(sz[0], sz[1], 1, 1, 1)

        scale = x * th.sigmoid(out)
        return scale


class SpatialLayer(nn.Module):
    def __init__(self, kernel_size=7, conv=None):
        super(SpatialLayer, self).__init__()
        assert kernel_size in (3, 7), 'kernel size must be 3 or 7'
        padding = 3 if kernel_size == 7 else 1
        self.spatial = conv(2, 1, kernel_size=kernel_size, padding=padding, bias=False)

    def forward(self, x):
        avg_out = th.mean(x, dim=1, keepdim=True)
        max_out, _ = th.max(x, dim=1, keepdim=True)
        out = th.cat([avg_out, max_out], dim=1)
        out = self.spatial(out)
        scale = x * th.sigmoid(out)
        return scale


class CBAM(nn.Module):
    def __init__(self, dim, reduction=16, kernel_size=7, conv=None):
        super(CBAM, self).__init__()
        self.ch = ChannelLayer(input_channels=dim, reduction_ratio=reduction)
        self.sp = SpatialLayer(kernel_size=kernel_size, conv=conv)

    def forward(self, x):
        out = self.ch(x)
        out = self.sp(out)
        return out



import warnings
import torch


def _flatten(sequence):
    flat = [p.contiguous().view(-1) for p in sequence]
    return torch.cat(flat) if len(flat) > 0 else torch.tensor([])


def _flatten_convert_none_to_zeros(sequence, like_sequence):
    flat = [
        p.contiguous().view(-1) if p is not None else torch.zeros_like(q).view(-1)
        for p, q in zip(sequence, like_sequence)
    ]
    return torch.cat(flat) if len(flat) > 0 else torch.tensor([])


def _possibly_nonzero(x):
    return isinstance(x, torch.Tensor) or x != 0


def _scaled_dot_product(scale, xs, ys):
    """Calculate a scaled, vector inner product between lists of Tensors."""
    # Using _possibly_nonzero lets us avoid wasted computation.
    return sum([(scale * x) * y for x, y in zip(xs, ys) if _possibly_nonzero(x) or _possibly_nonzero(y)])


def _dot_product(xs, ys):
    """Calculate the vector inner product between two lists of Tensors."""
    return sum([x * y for x, y in zip(xs, ys)])


def _has_converged(y0, y1, rtol, atol):
    """Checks that each element is within the error tolerance."""
    error_tol = tuple(atol + rtol * torch.max(torch.abs(y0_), torch.abs(y1_)) for y0_, y1_ in zip(y0, y1))
    error = tuple(torch.abs(y0_ - y1_) for y0_, y1_ in zip(y0, y1))
    return all((error_ < error_tol_).all() for error_, error_tol_ in zip(error, error_tol))


def _convert_to_tensor(a, dtype=None, device=None):
    if not isinstance(a, torch.Tensor):
        a = torch.tensor(a)
    if dtype is not None:
        a = a.type(dtype)
    if device is not None:
        a = a.to(device)
    return a


def _is_finite(tensor):
    _check = (tensor == float('inf')) + (tensor == float('-inf')) + torch.isnan(tensor)
    return not _check.any()


def _decreasing(t):
    return (t[1:] < t[:-1]).all()


def _assert_increasing(t):
    assert (t[1:] > t[:-1]).all(), 't must be strictly increasing or decreasing'


def _is_iterable(inputs):
    try:
        iter(inputs)
        return True
    except TypeError:
        return False


def _norm(x):
    """Compute RMS norm."""
    if torch.is_tensor(x):
        return x.norm() / (x.numel()**0.5)
    else:
        return torch.sqrt(sum(x_.norm()**2 for x_ in x) / sum(x_.numel() for x_ in x))


def _handle_unused_kwargs(solver, unused_kwargs):
    if len(unused_kwargs) > 0:
        warnings.warn('{}: Unexpected arguments {}'.format(solver.__class__.__name__, unused_kwargs))


def _select_initial_step(fun, t0, y0, order, rtol, atol, f0=None):
    """Empirically select a good initial step.

    The algorithm is described in [1]_.

    Parameters
    ----------
    fun : callable
        Right-hand side of the system.
    t0 : float
        Initial value of the independent variable.
    y0 : ndarray, shape (n,)
        Initial value of the dependent variable.
    direction : float
        Integration direction.
    order : float
        Method order.
    rtol : float
        Desired relative tolerance.
    atol : float
        Desired absolute tolerance.

    Returns
    -------
    h_abs : float
        Absolute value of the suggested initial step.

    References
    ----------
    .. [1] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential
           Equations I: Nonstiff Problems", Sec. II.4.
    """
    t0 = t0.to(y0[0])
    if f0 is None:
        f0 = fun(t0, y0)

    rtol = rtol if _is_iterable(rtol) else [rtol] * len(y0)
    atol = atol if _is_iterable(atol) else [atol] * len(y0)

    scale = tuple(atol_ + torch.abs(y0_) * rtol_ for y0_, atol_, rtol_ in zip(y0, atol, rtol))

    d0 = tuple(_norm(y0_ / scale_) for y0_, scale_ in zip(y0, scale))
    d1 = tuple(_norm(f0_ / scale_) for f0_, scale_ in zip(f0, scale))

    if max(d0).item() < 1e-5 or max(d1).item() < 1e-5:
        h0 = torch.tensor(1e-6).to(t0)
    else:
        h0 = 0.01 * max(d0_ / d1_ for d0_, d1_ in zip(d0, d1))

    y1 = tuple(y0_ + h0 * f0_ for y0_, f0_ in zip(y0, f0))
    f1 = fun(t0 + h0, y1)

    d2 = tuple(_norm((f1_ - f0_) / scale_) / h0 for f1_, f0_, scale_ in zip(f1, f0, scale))

    if max(d1).item() <= 1e-15 and max(d2).item() <= 1e-15:
        h1 = torch.max(torch.tensor(1e-6).to(h0), h0 * 1e-3)
    else:
        h1 = (0.01 / max(d1 + d2))**(1. / float(order + 1))

    return torch.min(100 * h0, h1)


def _compute_error_ratio(error_estimate, error_tol=None, rtol=None, atol=None, y0=None, y1=None):
    if error_tol is None:
        assert rtol is not None and atol is not None and y0 is not None and y1 is not None
        rtol if _is_iterable(rtol) else [rtol] * len(y0)
        atol if _is_iterable(atol) else [atol] * len(y0)
        error_tol = tuple(
            atol_ + rtol_ * torch.max(torch.abs(y0_), torch.abs(y1_))
            for atol_, rtol_, y0_, y1_ in zip(atol, rtol, y0, y1)
        )
    error_ratio = tuple(error_estimate_ / error_tol_ for error_estimate_, error_tol_ in zip(error_estimate, error_tol))
    mean_sq_error_ratio = tuple(torch.mean(error_ratio_ * error_ratio_) for error_ratio_ in error_ratio)
    return mean_sq_error_ratio


def _optimal_step_size(last_step, mean_error_ratio, safety=0.9, ifactor=10.0, dfactor=0.2, order=5):
    """Calculate the optimal size for the next step."""
    mean_error_ratio = max(mean_error_ratio)  # Compute step size based on highest ratio.
    if mean_error_ratio == 0:
        return last_step * ifactor
    if mean_error_ratio < 1:
        dfactor = _convert_to_tensor(1, dtype=torch.float64, device=mean_error_ratio.device)
    error_ratio = torch.sqrt(mean_error_ratio).to(last_step)
    exponent = torch.tensor(1 / order).to(last_step)
    factor = torch.max(1 / ifactor, torch.min(error_ratio**exponent / safety, 1 / dfactor))
    return last_step / factor


def _check_inputs(func, y0, t):
    tensor_input = False
    if torch.is_tensor(y0):
        tensor_input = True
        y0 = (y0,)
        _base_nontuple_func_ = func
        func = lambda t, y: (_base_nontuple_func_(t, y[0]),)
    assert isinstance(y0, tuple), 'y0 must be either a torch.Tensor or a tuple'
    for y0_ in y0:
        assert torch.is_tensor(y0_), 'each element must be a torch.Tensor but received {}'.format(type(y0_))

    if _decreasing(t):
        t = -t
        _base_reverse_func = func
        func = lambda t, y: tuple(-f_ for f_ in _base_reverse_func(-t, y))

    for y0_ in y0:
        if not torch.is_floating_point(y0_):
            raise TypeError('`y0` must be a floating point Tensor but is a {}'.format(y0_.type()))
    if not torch.is_floating_point(t):
        raise TypeError('`t` must be a floating point Tensor but is a {}'.format(t.type()))

    return tensor_input, func, y0, t
